/*
 * Tkinter 3000 -- Widget Construction Kit
 * $Id$
 *
 * Simple Tk-based drawing layer.  This implementation should run
 * on any platform supported by Tk.
 *
 * history:
 * 1998-09-13 fl   created (Tk drawing layer)
 * 2000-12-30 fl   adapted to Tkinter 3000
 * 2000-12-31 fl   added more special cases to getpoints
 * 2000-01-07 fl   added support for pixmap drawables
 * 2000-01-10 fl   more drawing primitives
 * 2000-01-11 fl   added setmargin; fixed paste offset; minor optimizations
 * 2000-02-04 fl   minor cleanup
 *
 * Copyright (c) 1998-2001 by Secret Labs AB
 * Copyright (c) 1998-2001 by Fredrik Lundh
 *
 * info@pythonware.com
 * http://www.pythonware.com
 */

/* --------------------------------------------------------------------
 * The Tkinter 3000 Widget Construction Kit is
 *
 * Copyright (c) 2000-2001 by Secret Labs AB
 * Copyright (c) 2000-2001 by Fredrik Lundh
 *
 * By obtaining, using, and/or copying this software and/or its
 * associated documentation, you agree that you have read, understood,
 * and will comply with the following terms and conditions:
 *
 * Permission to use, copy, modify, and distribute this software and its
 * associated documentation for any purpose and without fee is hereby
 * granted, provided that the above copyright notice appears in all
 * copies, and that both that copyright notice and this permission notice
 * appear in supporting documentation, and that the name of Secret Labs
 * AB or the author not be used in advertising or publicity pertaining to
 * distribution of the software without specific, written prior
 * permission.
 *
 * SECRET LABS AB AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO
 * THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS.  IN NO EVENT SHALL SECRET LABS AB OR THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
 * OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 * -------------------------------------------------------------------- */

#include <stdio.h>

#include "tk.h"
/* #include "tkInt.h" */

#include "Python.h"

/* -------------------------------------------------------------------- */
/* window drawing context */

typedef struct {
    PyObject_HEAD
    /* attributes */
    Tcl_Interp* interp;
    Tk_Window window;
    Display* display;
    Drawable drawable;
    int xoffset, yoffset;
    unsigned int is_pixmap:1;
} DrawObject;

staticforward PyTypeObject Draw_Type;

#define Draw_Check(op) ((op) != NULL && (op)->ob_type == &Draw_Type)

#define Draw_GetDrawable(self)\
    (self->drawable != None) ? self->drawable :\
    (self->drawable = Tk_WindowId(self->window))

/* -------------------------------------------------------------------- */
/* tk pen objects */

typedef struct {
    PyObject_HEAD
    /* attributes */
    GC gc;
    XColor *color;
    int width;
    Display* display;
} PenObject;

staticforward PyTypeObject Pen_Type;

#define Pen_Check(op) ((op) != NULL && (op)->ob_type == &Pen_Type)
#define Pen_GetGC(op, drawable) (((PenObject*)(op))->gc)

/* -------------------------------------------------------------------- */
/* tk brush objects */

typedef struct {
    PyObject_HEAD
    /* attributes */
    GC gc;
    XColor *color;
    Display* display;
} BrushObject;

staticforward PyTypeObject Brush_Type;

#define Brush_Check(op) ((op) != NULL && (op)->ob_type == &Brush_Type)
#define Brush_GetGC(op, drawable)\
 ((((BrushObject*)(op))->gc != None) ? (((BrushObject*)(op))->gc) :\
 Tk_GCForColor((((BrushObject*)(op))->color), (drawable)))

/* -------------------------------------------------------------------- */
/* tk font objects */

typedef struct {
    PyObject_HEAD
    /* attributes */
    GC gc;
    XColor *color;
    Tk_Font tkfont;
    int ascent;
    int descent;
    int height;
    Display* display;
} FontObject;

staticforward PyTypeObject Font_Type;

#define Font_Check(op) ((op) != NULL && (op)->ob_type == &Font_Type)
#define Font_GetGC(op, drawable) (((FontObject*)(op))->gc)

/* -------------------------------------------------------------------- */
/* helpers */

/* convert a color description to an RGB mask */
#define XCOLOR_RGB(xc)\
    (((xc)->red&0xff00)<<8) + ((xc)->green&0xff00) + (((xc)->blue&0xff00)>>8)

/* pick the pen from a set of drawing objects (NULL if neither is a pen) */
#define PEN(p1, p2)\
    (PenObject*) (Pen_Check(p1) ? (p1) : (Pen_Check(p2) ? (p2) : NULL))

/* pick the brush from a set of drawing objects (NULL if neither is a brush) */
#define BRUSH(p1, p2)\
    (BrushObject*) (Brush_Check(p1) ? (p1) : (Brush_Check(p2) ? (p2) : NULL))

static int
_lookupcolor(DrawObject* self, PyObject* colorIn)
{
    XColor* xc;
    int color;

    if (PyInt_Check(colorIn)) {
        /* integer: 0xrrggbb */
        color = (int) PyInt_AS_LONG(colorIn);

    } else if (PyString_Check(colorIn)) {
        /* string: lookup in colour database */
        xc = Tk_GetColor(self->interp, self->window,
                         PyString_AS_STRING(colorIn));
        if (!xc) {
            PyErr_SetString(PyExc_ValueError, "unknown color name");
            return -1;
        }
        color = XCOLOR_RGB(xc);
        Tk_FreeColor(xc);

    } else {
        PyErr_SetString(PyExc_TypeError, "expected string or integer");
        return -1;
    }

    return color;
}

/* -------------------------------------------------------------------- */

static PenObject *
draw_getpen(DrawObject* self, PyObject* args)
{
    XColor xc;
    XGCValues xgc;
    PenObject* pen;
    int color;

    PyObject* colorIn;
    int width;
    if (!PyArg_ParseTuple(args, "Oi:getpen", &colorIn, &width))
        return NULL;

    color = _lookupcolor(self, colorIn);
    if (color == -1)
        return NULL;

    pen = PyObject_NEW(PenObject, &Pen_Type);
    if (!pen)
        return NULL;

    pen->width = width;
    pen->display = self->display;

    xc.red = (color>>8) & 0xff00;
    xc.green = color & 0xff00;
    xc.blue = (color<<8) & 0xff00;
    pen->color = Tk_GetColorByValue(self->window, &xc);
    if (!pen->color) {
        PyErr_SetString(PyExc_ValueError, "unknown color name");
        return 0;
    }

    xgc.foreground = pen->color->pixel;
    xgc.line_width = width;

    pen->gc = Tk_GetGC(self->window, GCForeground|GCLineWidth, &xgc);

    return pen;
}

static void
pen_dealloc(PenObject* pen)
{
    Tk_FreeGC(pen->display, pen->gc);
    Tk_FreeColor(pen->color);

    PyMem_DEL(pen);
}

static PyMethodDef pen_methods[] = {
    {NULL, NULL}
};

static PyObject*  
pen_getattr(PenObject* pen, char* name)
{
    return Py_FindMethod(pen_methods, (PyObject*) pen, name);
}

statichere PyTypeObject Pen_Type = {
    PyObject_HEAD_INIT(NULL)
    0, "Pen", sizeof(PenObject), 0,
    (destructor)pen_dealloc, /* tp_dealloc */
    0, /* tp_print */
    (getattrfunc)pen_getattr, /* tp_getattr */
};

/* -------------------------------------------------------------------- */

static BrushObject *
draw_getbrush(DrawObject* self, PyObject* args)
{
    XColor xc;
    BrushObject* brush;
    int color;

    PyObject* colorIn;
    if (!PyArg_ParseTuple(args, "O:getbrush", &colorIn))
        return NULL;

    color = _lookupcolor(self, colorIn);
    if (color == -1)
        return NULL;

    brush = PyObject_NEW(BrushObject, &Brush_Type);
    if (!brush)
        return NULL;

    brush->display = self->display;

    xc.red = (color>>8) & 0xff00;
    xc.green = color & 0xff00;
    xc.blue = (color<<8) & 0xff00;
    brush->color = Tk_GetColorByValue(self->window, &xc);
    if (!brush->color) {
        PyErr_SetString(PyExc_ValueError, "unknown color name");
        return 0;
    }

    brush->gc = None;

    return brush;
}

static void
brush_dealloc(BrushObject* brush)
{
    if (brush->gc)
        Tk_FreeGC(brush->display, brush->gc);

    Tk_FreeColor(brush->color);

    PyMem_DEL(brush);
}

static PyMethodDef brush_methods[] = {
    {NULL, NULL}
};

static PyObject*  
brush_getattr(BrushObject* brush, char* name)
{
    return Py_FindMethod(brush_methods, (PyObject*) brush, name);
}

statichere PyTypeObject Brush_Type = {
    PyObject_HEAD_INIT(NULL)
    0, "Brush", sizeof(BrushObject), 0,
    (destructor)brush_dealloc, /* tp_dealloc */
    0, /* tp_print */
    (getattrfunc)brush_getattr, /* tp_getattr */
};

/* -------------------------------------------------------------------- */

static PyObject *
draw_getfont(DrawObject* self, PyObject* args)
{
    XColor xc;
    XGCValues xgc;
    FontObject* font;
    Tk_FontMetrics fm;
    int color;

    PyObject* colorIn;
    char* fontname;
    if (!PyArg_ParseTuple(args, "Os:getfont", &colorIn, &fontname))
        return NULL;

    color = _lookupcolor(self, colorIn);
    if (color == -1)
        return NULL;

    font = PyObject_NEW(FontObject, &Font_Type);
    if (!font)
	return NULL;

    font->display = self->display;

    xc.red = (color>>8) & 0xff00;
    xc.green = color & 0xff00;
    xc.blue = (color<<8) & 0xff00;
    font->color = Tk_GetColorByValue(self->window, &xc);
    if (!font->color) {
        PyErr_SetString(PyExc_ValueError, "unknown color");
        PyMem_DEL(font);
        return 0;
    }

    font->tkfont = Tk_GetFont(self->interp, self->window, fontname);
    if (!font->tkfont) {
        Tk_FreeColor(font->color);
        PyErr_SetString(PyExc_ValueError, "unknown font name");
        PyMem_DEL(font);
        return NULL;
    }

    /* get basic font metrics */
    Tk_GetFontMetrics(font->tkfont, &fm);
    font->ascent = fm.ascent;
    font->descent = fm.descent;
    font->height = fm.ascent + fm.descent;

    xgc.foreground = font->color->pixel;
    xgc.font = Tk_FontId(font->tkfont);

    font->gc = Tk_GetGC(self->window, GCForeground|GCFont, &xgc);

    return (PyObject*) font;
}

static void
font_dealloc(FontObject* font)
{
    Tk_FreeGC(font->display, font->gc);
    Tk_FreeColor(font->color);
    Tk_FreeFont(font->tkfont);

    PyMem_DEL(font);
}

static PyObject *
font_measure(FontObject* font, PyObject* args)
{
    int width;

    PyObject* string = NULL;
    if (!PyArg_ParseTuple(args, "|O!:measure", &PyString_Type, &string))
        return NULL;

    if (string)
        width = Tk_TextWidth(
            font->tkfont,
            PyString_AS_STRING(string), PyString_GET_SIZE(string)
            );
    else {
        /* return average size */
        width = Tk_TextWidth(
            font->tkfont,
            "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz", 52
            );
        width = (width + 26) / 52;
    }

    return Py_BuildValue("ii", width, font->height);
}

static PyMethodDef font_methods[] = {
    {"measure", (PyCFunction)font_measure, 1},
    {NULL, NULL}
};

static PyObject*  
font_getattr(FontObject* font, char* name)
{
    return Py_FindMethod(font_methods, (PyObject*) font, name);
}

statichere PyTypeObject Font_Type = {
    PyObject_HEAD_INIT(NULL)
    0, "Font", sizeof(FontObject), 0,
    (destructor)font_dealloc, /* tp_dealloc */
    0, /* tp_print */
    (getattrfunc)font_getattr, /* tp_getattr */
};

/* -------------------------------------------------------------------- */

static PyObject*
draw_lookupcolor(DrawObject* self, PyObject* args)
{
    /* map colour name to RGB value */

    int color;

    PyObject* colorIn;
    if (!PyArg_ParseTuple(args, "O:_lookupcolor", &colorIn))
        return NULL;

    color = _lookupcolor(self, colorIn);
    if (color == -1)
        return NULL;

    return Py_BuildValue("i", color);
}

static PyObject*
draw_lookuprelief(DrawObject* self, PyObject* args)
{
    /* map relief name to Tk relief code */

    int relief;
    int ok;

    char* reliefIn;
    if (!PyArg_ParseTuple(args, "s:_lookuprelief", &reliefIn))
        return NULL;

    ok = Tk_GetRelief(self->interp, reliefIn, &relief);
    if (ok != TCL_OK) {
        PyErr_SetString(PyExc_TypeError, "unrecognized relief value");
        return NULL;
    }

    return Py_BuildValue("i", relief);
}

static PyObject*
draw_setmargin(DrawObject* self, PyObject* args)
{
    /* (re)set display offset */

    int margin;
    if (!PyArg_ParseTuple(args, "i:_setoffset", &margin))
        return NULL;

    self->xoffset = self->yoffset = margin;

    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject*
draw_getpixmap(DrawObject* self, PyObject* args)
{
    /* create compatible pixmap for the given drawable */
    DrawObject* draw;

    int width, height;
    if (!PyArg_ParseTuple(args, "ii:getpixmap", &width, &height))
        return NULL;

    draw = PyObject_NEW(DrawObject, &Draw_Type);
    if (!draw)
	return NULL;

    draw->is_pixmap = 1;

    draw->interp = self->interp;

    draw->display = self->display;

    draw->window = self->window;

    /* FIXME: add more error checking here */
    draw->drawable = Tk_GetPixmap(
        self->display, Tk_WindowId(self->window),
        width, height, Tk_Depth(self->window)
        );

    draw->xoffset = draw->yoffset = 0;

    return (PyObject*) draw;
}

static PyObject* 
module_getdraw(PyObject* self, PyObject* args)
{
    DrawObject* draw;

    long arg;
    char* name;
    if (!PyArg_ParseTuple(args, "ls:getdraw", &arg, &name))
        return NULL;

    draw = PyObject_NEW(DrawObject, &Draw_Type);
    if (!draw)
	return NULL;

    draw->is_pixmap = 0;

    draw->interp = (Tcl_Interp*) arg;

    /* FIXME: add more error checking here */
    draw->window = Tk_NameToWindow(
        draw->interp, name,
        Tk_MainWindow(draw->interp)
        );

    draw->display = Tk_Display(draw->window);

    draw->drawable = None;

    draw->xoffset = draw->yoffset = 0;

    return (PyObject*) draw;
}

static void
draw_dealloc(DrawObject* self)
{
    if (self->is_pixmap)
        Tk_FreePixmap(self->display, self->drawable);
    PyMem_DEL(self);
}

/* -------------------------------------------------------------------- */
/* drawing primitives */

#define GETINT(op)\
    (PyInt_Check(op) ? PyInt_AS_LONG((op)) :\
     PyFloat_Check(op) ? (int) PyFloat_AS_DOUBLE((op)) :\
     PyInt_AsLong(op))

static XPoint*
getpoints(DrawObject* self, PyObject* xyIn, int* count, int auto_close)
{
    XPoint *xy;
    int i, n;
    int x, y;

    if (!PySequence_Check(xyIn)) {
        PyErr_SetString(PyExc_TypeError, "argument must be a sequence");
        return NULL;
    }

    n = PyObject_Length(xyIn);
    if (PyErr_Occurred())
        return NULL;

    if (n & 1) {
        PyErr_SetString(PyExc_TypeError,
                        "expected even number of coordinates");
        return NULL;
        }

    n /= 2;
    xy = malloc((n+1) * sizeof(XPoint));
    if (!xy) {
        PyErr_NoMemory();
        *count = -1;
        return NULL;
    }

    x = self->xoffset;
    y = self->yoffset;

    if (PyList_Check(xyIn))
        for (i = 0; i < n; i++) {
            xy[i].x = (short) (GETINT(PyList_GET_ITEM(xyIn, i+i)) + x);
            xy[i].y = (short) (GETINT(PyList_GET_ITEM(xyIn, i+i+1)) + y);
        }
    else if (PyTuple_Check(xyIn))
        for (i = 0; i < n; i++) {
            xy[i].x = (short) (GETINT(PyTuple_GET_ITEM(xyIn, i+i)) + x);
            xy[i].y = (short) (GETINT(PyTuple_GET_ITEM(xyIn, i+i+1)) + y);
        }
    else
        for (i = 0; i < n; i++) {
            PyObject *op;
            op = PySequence_GetItem(xyIn, i+i);
            xy[i].x = (short) (GETINT(op) + x);
            Py_DECREF(op);
            op = PySequence_GetItem(xyIn, i+i+1);
            xy[i].y = (short) (GETINT(op) + y);
            Py_DECREF(op);
        }

    PyErr_Clear();

    /* close polygon */
    if (auto_close && i > 0 && (xy[i-1].x != xy[0].x ||
                                xy[i-1].y != xy[0].y)) {
        xy[i].x = xy[0].x;
        xy[i].y = xy[0].y;
        n++;
    }

    *count = n;

    return xy;
}

/* -------------------------------------------------------------------- */


static PyObject* 
draw_ellipse(DrawObject* self, PyObject* args)
{
    Drawable drawable;
    PyObject* resource;

    PyObject* pen = NULL;
    PyObject* brush = NULL;
    int x0, y0, x1, y1;
    if (!PyArg_ParseTuple(args, "(iiii)|OO:ellipse", &x0, &y0, &x1, &y1,
                          &pen, &brush))
        return NULL;

    x0 = x0 + self->xoffset; x1 = x1 + self->xoffset;
    y0 = y0 + self->yoffset; y1 = y1 + self->yoffset;

    drawable = Draw_GetDrawable(self);

    resource = (PyObject*) BRUSH(brush, pen);
    if (resource)
        XFillArc(
            self->display, drawable,
	    Brush_GetGC(resource, drawable),
            x0, y0, (unsigned) (x1 - x0 - 1), (unsigned) (y1 - y0 - 1),
            0, 360*64);

    resource = (PyObject*) PEN(pen, brush);
    if (resource)
        XDrawArc(
            self->display, drawable,
	    Pen_GetGC(resource, drawable),
            x0, y0, (unsigned) (x1 - x0 - 1), (unsigned) (y1 - y0 - 1),
            0, 360*64);

    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject* 
draw_line(DrawObject* self, PyObject* args)
{
    XPoint* xy;
    int count;
    Drawable drawable;

    PyObject* xyIn;
    PyObject* pen = NULL;
    if (!PyArg_ParseTuple(args, "O|O:line", &xyIn, &pen))
            return NULL;

    if (Pen_Check(pen)) {

        xy = getpoints(self, xyIn, &count, 0);
        if (!xy)
            return NULL;

        drawable = Draw_GetDrawable(self);

        XDrawLines(self->display, drawable,
                   Pen_GetGC(pen, drawable),
                   xy, count, CoordModeOrigin);

        free(xy);

    }

    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject* 
draw_polygon(DrawObject* self, PyObject* args)
{
    XPoint* xy;
    int count;
    Drawable drawable;
    PyObject* resource;

    PyObject* xyIn;
    PyObject* pen = NULL;
    PyObject* brush = NULL;
    if (!PyArg_ParseTuple(args, "O|OO:polygon", &xyIn, &pen, &brush))
        return NULL;

    xy = getpoints(self, xyIn, &count, 1);
    if (!xy)
        return NULL;

    drawable = Draw_GetDrawable(self);

    resource = (PyObject*) BRUSH(brush, pen);
    if (resource)
        XFillPolygon(
            self->display, drawable,
	    Brush_GetGC(resource, drawable),
            xy, count-1, Complex, CoordModeOrigin);

    resource = (PyObject*) PEN(pen, brush);
    if (resource)
        XDrawLines(
            self->display, drawable,
	    Pen_GetGC(resource, drawable),
            xy, count, CoordModeOrigin);

    free(xy);

    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject* 
draw_rectangle(DrawObject* self, PyObject* args)
{
    Drawable drawable;
    PyObject* resource;

    PyObject* pen = NULL;
    PyObject* brush = NULL;
    int x0, y0, x1, y1;
    if (!PyArg_ParseTuple(args, "(iiii)|OO:rectangle", &x0, &y0, &x1, &y1,
                          &pen, &brush))
        return NULL;

    x0 = x0 + self->xoffset; x1 = x1 + self->xoffset;
    y0 = y0 + self->yoffset; y1 = y1 + self->yoffset;

    drawable = Draw_GetDrawable(self);

    resource = (PyObject*) BRUSH(brush, pen);
    if (resource)
        XFillRectangle(
            self->display, drawable,
	    Brush_GetGC(resource, drawable),
            x0, y0, (unsigned) (x1 - x0), (unsigned) (y1 - y0));

    resource = (PyObject*) PEN(pen, brush);
    if (resource)
        XDrawRectangle(
            self->display, drawable,
	    Pen_GetGC(resource, drawable),
            x0, y0, (unsigned) (x1 - x0), (unsigned) (y1 - y0));

    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject* 
draw_text(DrawObject* self, PyObject* args)
{
    Drawable drawable;

    PyObject* text;
    FontObject* font;
    int x0, y0;
    if (!PyArg_ParseTuple(args, "(ii)|O!O!:text", &x0, &y0,
                          &PyString_Type, &text,
                          &Font_Type, &font))
        return NULL;

    x0 = x0 + self->xoffset;
    y0 = y0 + self->yoffset;

    drawable = Draw_GetDrawable(self);

    Tk_DrawChars(
        self->display, drawable,
        Font_GetGC(font, drawable),
        font->tkfont, PyString_AS_STRING(text), PyString_GET_SIZE(text),
        x0, y0 + font->ascent
        );

    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject* 
draw_paste(DrawObject* self, PyObject* args)
{
    GC gc;

    /* FIXME: experimental interface.  may change in future releases */
    /* change coordinate arguments to match PIL (add crop object?) */

    DrawObject* draw;
    int width = -1, height = -1;
    int xoffset = 0, yoffset = 0;
    if (!PyArg_ParseTuple(args, "O!|(ii)(ii):paste", &Draw_Type, &draw,
                          &width, &height, &xoffset, &yoffset))
        return NULL;

    xoffset += self->xoffset;
    yoffset += self->yoffset;

    if (width < 0)
        width = Tk_Width(draw->window);
    if (height < 0)
        height = Tk_Height(draw->window);

    gc = Tk_GetGC(self->window, 0, NULL);

    XCopyArea(self->display, Draw_GetDrawable(draw),
              Draw_GetDrawable(self), gc,
              0, 0, width, height, xoffset, yoffset);

    Tk_FreeGC(self->display, gc);

    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject* 
draw_border(DrawObject* self, PyObject* args)
{
    Tk_3DBorder border3D;
    Tk_Uid uid;
    char color[16];

    /* internal interface.  may change in future releases */

    BrushObject* brush;
    int margin, border, relief;
    if (!PyArg_ParseTuple(args, "O!iii:_border",
                          &Brush_Type, &brush,
                          &margin, &border, &relief))
        return NULL;

    if (margin + border <= 0) {
        Py_INCREF(Py_None);
        return Py_None;
    }

    sprintf(color, "#%04x%04x%04x", brush->color->red,
            brush->color->green, brush->color->blue);

    uid = Tk_GetUid(color);

    border3D = Tk_Get3DBorder(self->interp, self->window, uid);

    Tk_Draw3DRectangle(
        self->window, Draw_GetDrawable(self), border3D,
        margin, margin, Tk_Width(self->window)-2*margin,
        Tk_Height(self->window)-2*margin,
        border, relief);

    Tk_Free3DBorder(border3D);

    Py_INCREF(Py_None);
    return Py_None;
}

static PyMethodDef draw_methods[] = {
    /* draw interface */
    {"text", (PyCFunction)draw_text, 1},
    {"line", (PyCFunction)draw_line, 1},
    {"rectangle", (PyCFunction)draw_rectangle, 1},
    {"polygon", (PyCFunction)draw_polygon, 1},
    {"paste", (PyCFunction)draw_paste, 1},
    {"ellipse", (PyCFunction)draw_ellipse, 1},
    /* factories (access via widget interface) */
    {"getpen", (PyCFunction)draw_getpen, 1},
    {"getbrush", (PyCFunction)draw_getbrush, 1},
    {"getfont", (PyCFunction)draw_getfont, 1},
    {"getpixmap", (PyCFunction)draw_getpixmap, 1},
    /* helpers (internal; subject to change) */
    {"_border", (PyCFunction)draw_border, 1},
    {"_setmargin", (PyCFunction)draw_setmargin, 1},
    {"_lookupcolor", (PyCFunction)draw_lookupcolor, 1},
    {"_lookuprelief", (PyCFunction)draw_lookuprelief, 1},
    {NULL, NULL}
};

static PyObject*  
draw_getattr(DrawObject* self, char* name)
{
    return Py_FindMethod(draw_methods, (PyObject*) self, name);
}

statichere PyTypeObject Draw_Type = {
    PyObject_HEAD_INIT(NULL)
    0, "Draw", sizeof(DrawObject), 0, 
    /* methods */
    (destructor)draw_dealloc, /*tp_dealloc*/
    0,                          /*tp_print*/
    (getattrfunc)draw_getattr, /*tp_getattr*/
};

/* -------------------------------------------------------------------- */

static PyMethodDef module_functions[] = {
    {"getdraw", (PyCFunction)module_getdraw, 1},
    {NULL, NULL}
};

void
#ifdef WIN32
__declspec(dllexport)
#endif
init_tk3draw()
{
    /* Patch object types */
    Pen_Type.ob_type = Brush_Type.ob_type =
        Draw_Type.ob_type = Font_Type.ob_type = &PyType_Type;

    Py_InitModule("_tk3draw", module_functions);
}
