#
# Tkinter 3000 -- Widget Construction Kit
# $Id$
#
# a pure Python implementation of the uiToolkit widget
# construction kit interface
#
# history:
# 2000-12-30 fl   created
# 2001-01-03 fl   first preview release
# 2001-01-05 fl   added controller support
# 2001-01-07 fl   added scroll mixin; rename internal mixin methods
# 2001-01-09 fl   experimental support for double-buffered rendering
# 2001-01-10 fl   added border support
# 2001-01-25 fl   minor tweaks; made destroy a bit more robust
# 2001-01-26 fl   added event controller/mixin (for simplied event handling)
# 2001-02-01 fl   changed init handling to better match uiToolkit
#
# Copyright (c) 2000-2001 by Secret Labs AB
# Copyright (c) 2000-2001 by Fredrik Lundh
#
# info@pythonware.com
# http://www.pythonware.com
#
# --------------------------------------------------------------------
# The Tkinter 3000 Widget Construction Kit is
#
# Copyright (c) 2000-2001 by Secret Labs AB
# Copyright (c) 2000-2001 by Fredrik Lundh
#
# By obtaining, using, and/or copying this software and/or its
# associated documentation, you agree that you have read, understood,
# and will comply with the following terms and conditions:
#
# Permission to use, copy, modify, and distribute this software and its
# associated documentation for any purpose and without fee is hereby
# granted, provided that the above copyright notice appears in all
# copies, and that both that copyright notice and this permission notice
# appear in supporting documentation, and that the name of Secret Labs
# AB or the author not be used in advertising or publicity pertaining to
# distribution of the software without specific, written prior
# permission.
#
# SECRET LABS AB AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO
# THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS.  IN NO EVENT SHALL SECRET LABS AB OR THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
# OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
# --------------------------------------------------------------------

# TODO: make ui_draw a private attribute ???
# TODO: bitmap/image support
# TODO: ui_handle_map ???
# TODO: distutils script
# TODO: a "wck tour" demo script

import Tkinter
import sys

# standard drawing engine
import _tk3draw

Error = Tkinter.TclError

if sys.platform == "win32":
    FONT = "{ms sans serif} 8"
    FOREGROUND = "systembuttontext"
    BACKGROUND = "systembuttonface"
else:
    FONT = "Helvetica 12"
    FOREGROUND = "black"
    BACKGROUND = "grey"

# used to hide inherited options
NOT_INHERITED = [None]

class Style:
    # FIXME: move this to _tk3draw

    margin = 0

    def __init__(self, widget):
        margin = max(int(widget.ui_option_highlightthickness), 0)
        border = max(int(widget.ui_option_borderwidth), 0)
        relief = widget.ui_draw._lookuprelief(str(widget.ui_option_relief))
        self.background = widget.ui_brush(widget.ui_option_background)
        if margin + border > 0:
            self.margin = margin + border
            self.border = self.background, margin, border, relief

    def getmargin(self):
        return self.margin

    def drawbackground(self, xy, draw):
        draw.rectangle(xy, self.background)

    def drawborder(self, xy, draw):
        if self.margin:
            apply(draw._border, self.border)

class Widget(Tkinter.Widget):

    ui_option_background = BACKGROUND

    ui_option_relief = "flat"
    ui_option_borderwidth = 0

    ui_option_highlightthickness = 0

    ui_option_cursor = ""

    ui_doublebuffer = 0 # experimental

    ui_controller = None

    def __init__(self, master, **options):
        # feel free to override __init__ in order to initialize
        # instance variables (etc).  just don't forget to call
        # ui_init when you're finished (or Widget.__init__, if
        # you prefer)

        self.ui_init(master, options)

    def ui_init(self, master, options):

        # create a frame widget
        Tkinter.Widget.__init__(
            self, master, "frame",
            {"background": "", "takefocus": ""}
            )

        # initialize the widget construction kit
        self.ui_draw = _tk3draw.getdraw(self.tk.interpaddr(), str(self))
        self.__size = 100, 100 # default size
        self.__style = None # default style
        self.__visible = 0
        self.__repair_pending = 0

        # attach controllers to widget
        default_controller = getcontroller(_DefaultController, self)
        widget_controller = getcontroller(self.ui_controller, self)
        if widget_controller:
            controllers = widget_controller, default_controller
        else:
            controllers = (default_controller,)
        tags = self.bindtags()
        self.bindtags(tags[:1] + controllers + tags[2:])

        # initial configure
        self.__configure(options, self.__size)

    def _ui_destroy(self):
        # internal
        if hasattr(self, "ui_draw"):
            try:
                self.ui_handle_destroy()
            finally:
                save = self.tk, self.master, self.children
                self.__dict__.clear()
                self.tk, self.master, self.children = save

    def destroy(self):
        "Destroy widget"
        try:
            Tkinter.Widget.destroy(self)
        finally:
            self._ui_destroy()

    #
    # option database

    def __configure(self, options, default_size=None):

        items = options.items()

        values = []
        try:
            for option, value in items:
                value = getattr(self, "ui_option_" + option)
                if value is NOT_INHERITED:
                    raise AttributeError
                values.append(value)
        except AttributeError:
            raise Error, "unsupported option: %s" % option

        for option, value in items:
            setattr(self, "ui_option_" + option, value)

        try:
            size = self.ui_handle_config()
            self.__style = Style(self)
        except:
            # rollback
            for option, value in items:
                setattr(self, "ui_option_" + option, values.pop(0))
            raise

        margin = self.__style.getmargin()
        self.ui_draw._setmargin(margin)

        if size or default_size:
            if not size:
                size = default_size
            margin = 2 * margin
            self.tk.call(
                self, "config",
                "-width", size[0] + margin, "-height", size[1] + margin
                )

        if options.has_key("cursor"):
            self.tk.call(
                self, "config",
                "-cursor", str(self.ui_option_cursor)
                )

        return size

    def configure(self, **options):
        "Set options"
        if not options:
            # FIXME
            raise NotYetImplemented, "configure query not yet implemented"
        self.ui_damage()
        return self.__configure(options)

    config = configure

    def cget(self, option):
        "Get option value"
        try:
            return getattr(self, "ui_option_" + option)
        except AttributeError:
            raise Error, "unsupported option: %s" % option

    __getitem__ = cget

    def __setitem__(self, option, value):
        self.configure({option: value})

    def keys(self):
        "Get available options"
        return self.configure().keys()

    #
    # construction kit

    def ui_handle_clear(self, draw, x0, y0, x1, y1):
        self.__style.drawbackground((x0, y0, x1, y1), draw)

    def ui_handle_config(self):
        pass

    def ui_handle_damage(self, x0, y0, x1, y1):
        pass

    def ui_handle_destroy(self):
        pass

    def ui_handle_focus(self, draw, has_focus):
        pass

    def ui_handle_repair(self, draw, x0, y0, x1, y1):
        pass

    def ui_handle_resize(self, width, height):
        pass

    def ui_handle_select(self):
        pass # NYI

    def ui_damage(self, x0=None, y0=None, x1=None, y1=None):
        "Report widget damage"
        if not self.__visible:
            return
        if x0 is None:
            x0, y0, x1, y1 = (0, 0) + self.__size
        elif isinstance(x0, type(())):
            x0, y0, x1, y1 = x0
        if self.__repair_pending:
            if x1 > x0 or y1 > x0:
                self.ui_handle_damage(x0, y0, x1, y1)
        else:
            self.__repair_pending = 1
            self.tk.call(
                "event", "generate", self, "<Expose>",
                "-x", x0, "-y", y0, "-width", x1-x0, "-height", y1-y0,
                "-when", "tail"
                )

    def ui_size(self):
        "Get current widget (view) size"
        return self.__size

    def _ui_border(self, draw):
        # internal: redraw border
        if self.__style:
            self.__style.drawborder(None, draw)

    def _ui_setsize(self, width, height):
        # internal: set size
        margin = 2 * self.__style.getmargin()
        self.__size = width - margin, height - margin
        return self.__size

    def _ui_setvisible(self, visible):
        # internal: set visibility flag
        self.__visible = visible

    def _ui_reset_repair_pending(self):
        # internal: reset repair pending flag
        self.__repair_pending = 0

    #
    # object factories

    def ui_pen(self, color="black", width=1, **options):
        "Create pen object"
        return self.ui_draw.getpen(color, width)

    def ui_brush(self, color="black", **options):
        "Create brush object"
        return self.ui_draw.getbrush(color)

    def ui_font(self, color="black", font="Courier", **options):
        "Create font object"
        return self.ui_draw.getfont(color, font)

    def ui_image(self, image=None, **options):
        "Create image object"
        return self.ui_draw.getimage(image) # NYI

    def ui_purge(self):
        "Clear resource cache"
        pass

# controller classes.  note that all Tkinter3000 widgets share a
# single dispatcher instance

class Controller:

    tag = None
    __created = 0

    def ui_gettag(self, widget):
        # internal: lazy instantation
        if not self.tag:
            self.tag = "Tkinter3000.Controller.%d" % id(self)
        if not self.__created:
            # local binding function
            def handle(event, function, tag=self.tag, widget=widget):
                widget.bind_class(tag, event, function)
            self.create(handle)
            self.__created = 1
        return self.tag

    def create(self, handle):
        raise NotImplementedError, "must override create"


class _DefaultController(Controller):

    tag = "Tkinter3000.DefaultController"

    def create(self, handle):
        handle("<Configure>", self.handle_configure)
        handle("<Destroy>", self.handle_destroy)
        handle("<Expose>", self.handle_expose)
        handle("<FocusIn>", self.handle_focusin)
        handle("<FocusOut>", self.handle_focusout)
        handle("<Map>", self.handle_map)
        handle("<Unmap>", self.handle_unmap)

    def handle_configure(self, event):
        widget = event.widget
        width, height = widget._ui_setsize(event.width, event.height)
        widget.ui_handle_resize(width, height)

    def handle_destroy(self, event):
        widget = event.widget
        if isinstance(widget, Widget):
            widget._ui_destroy()

    def handle_expose(self, event):
        widget = event.widget
        x0 = event.x; y0 = event.y
        if event.width > 0 or event.height > 0:
            widget.ui_handle_damage(x0, y0, x0+event.width, y0+event.height)
## FIXME: event.count is not yet supported by Tkinter
##         try:
##             count = event.count
##         except AttributeError:
##             count = 0
##        if not count:
        draw = widget.ui_draw
        width, height = widget.ui_size()
        widget._ui_reset_repair_pending()
        if widget.ui_doublebuffer:
            pixmap = draw.getpixmap(width, height)
            widget.ui_handle_clear(pixmap, 0, 0, width, height)
            widget.ui_handle_repair(pixmap, 0, 0, width, height)
            widget._ui_border(pixmap)
            draw.paste(pixmap)
        else:
            widget.ui_handle_clear(draw, 0, 0, width, height)
            widget.ui_handle_repair(draw, 0, 0, width, height)
            widget._ui_border(draw)

    def handle_focusin(self, event):
        widget = event.widget
        widget.ui_handle_focus(widget.ui_draw, 1)

    def handle_focusout(self, event):
        widget = event.widget
        widget.ui_handle_focus(widget.ui_draw, 0)

    def handle_map(self, event):
        widget = event.widget
        widget._ui_setvisible(1)

    def handle_unmap(self, event):
        widget = event.widget
        widget._ui_setvisible(0)

_controller_registry = {}

def getcontroller(controller_factory, widget):
    "Create a controller for the given widget"
    try:
        controller =_controller_registry[controller_factory]
    except KeyError:
        if not callable(controller_factory):
            return None # no controller
        controller = controller_factory()
        _controller_registry[controller.tag] = controller
        _controller_registry[controller_factory] = controller
    return controller.ui_gettag(widget)

# --------------------------------------------------------------------
# make utility classes available from this module

from tk3utils import *
