#
# Tkinter 3000 -- Widget Construction Kit
# $Id$
#
# widget helpers
#
# history:
# 2001-01-07 fl   added scroll mixin; rename internal mixin methods
# 2001-01-26 fl   added event controller/mixin (for simplied event handling)
# 2001-02-03 fl   moved from tk3.py
#
# Copyright (c) 2000-2001 by Secret Labs AB
# Copyright (c) 2000-2001 by Fredrik Lundh
#
# info@pythonware.com
# http://www.pythonware.com
#
# --------------------------------------------------------------------
# The Tkinter 3000 Widget Construction Kit is
#
# Copyright (c) 2000-2001 by Secret Labs AB
# Copyright (c) 2000-2001 by Fredrik Lundh
#
# By obtaining, using, and/or copying this software and/or its
# associated documentation, you agree that you have read, understood,
# and will comply with the following terms and conditions:
#
# Permission to use, copy, modify, and distribute this software and its
# associated documentation for any purpose and without fee is hereby
# granted, provided that the above copyright notice appears in all
# copies, and that both that copyright notice and this permission notice
# appear in supporting documentation, and that the name of Secret Labs
# AB or the author not be used in advertising or publicity pertaining to
# distribution of the software without specific, written prior
# permission.
#
# SECRET LABS AB AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO
# THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS.  IN NO EVENT SHALL SECRET LABS AB OR THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
# OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
# --------------------------------------------------------------------

import tk3

# --------------------------------------------------------------------
# text mixin.  defines a font resource attribute, and handles width
# and height in character units.

class TextMixin:

    ui_option_foreground = tk3.FOREGROUND

    ui_option_font = tk3.FONT

    ui_option_height = 0
    ui_option_width = 0

    def ui_handle_config(self):
        # fonts
        self.font = font = self.ui_font(
            self.ui_option_foreground,
            self.ui_option_font
            )
        # size (default is actual size)
        option_width = int(self.ui_option_width)
        option_height = int(self.ui_option_height)
        if option_width <= 0 or option_height <= 0:
            try:
                width, height = font.measure(self.ui_text_get())
            except NotImplementedError:
                raise Error, "width and height must both be non-zero"
            if option_width > 0:
                width = font.measure()[0] * option_width
            if option_height > 0:
                height = height * option_height
        else:
            width, height = font.measure()
            width = option_width * width
            height = option_height * height
        return width, height

    def ui_text_draw(self, draw, text=None):
        # draw centered text
        size = self.ui_size()
        if text is None:
            text = self.ui_text_get()
        width, height = self.font.measure(text)
        draw.text(((size[0]-width)/2, (size[1]-height)/2), text, self.font)

    def ui_text_get(self):
        raise NotImplementedError, "must override text_get"

# --------------------------------------------------------------------
# event mixin.  simplifies user event handling, by mapping mouse and
# keyboard events to DOM/DHTML-style method calls

class EventController(tk3.Controller):

    def create(self, handle):
        handle("<Any-Key>", self.handle_key)
        handle("<Any-Button>", self.handle_press)
        handle("<Any-ButtonRelease>", self.handle_release)
        handle("<Enter>", self.handle_enter)
        handle("<Motion>", self.handle_motion)
        handle("<Leave>", self.handle_leave)

    def handle_key(self, event):
        event.widget.onkey(event)

    def handle_press(self, event):
        event.widget.onmousedown(event)

    def handle_release(self, event):
        # FIXME: only generate click event if the mouse hasn't moved
        event.widget.onclick(event)
        event.widget.onmouseup(event)

    def handle_enter(self, event):
        event.widget.onmouseover(event)

    def handle_motion(self, event):
        event.widget.onmousemove(event)

    def handle_leave(self, event):
        event.widget.onmouseout(event)

class EventMixin:

    ui_controller = EventController

    def onkey(self, event):
        pass

    def onclick(self, event):
        pass

    def onmousedown(self, event):
        pass

    def onmouseup(self, event):
        pass

    def onmouseover(self, event):
        pass

    def onmousemove(self, event):
        pass

    def onmouseout(self, event):
        pass

# --------------------------------------------------------------------
# button mixin.  implements basic button widget behaviour (arm/click)

class ButtonController(tk3.Controller):

    active = None
    inside = None

    def create(self, handle):
        handle("<Enter>", self.handle_enter)
        handle("<Leave>", self.handle_leave)
        handle("<Button-1>", self.handle_button_1)
        handle("<ButtonRelease-1>", self.handle_button_release_1)
        handle("<space>", self.handle_space)

    def handle_enter(self, event):
        widget = event.widget
        widget.ui_button_enter()
        if widget.cget("state") != "disabled":
            if widget is self.active:
                # arm if user moves pointer back into active button
                widget.ui_button_arm()
        self.inside = widget

    def handle_leave(self, event):
        widget = event.widget
        widget.ui_button_leave()
        if widget is self.active:
            # disarm button if user moves the pointer outside
            # the widget
            widget.ui_button_disarm()
        self.inside = None

    def handle_button_1(self, event):
        widget = event.widget
        if widget.cget("state") != "disabled":
            widget.ui_button_arm()
            self.active = widget

    def handle_button_release_1(self, event):
        widget = event.widget
        if self.active is widget and widget.cget("state") != "disabled":
            self.active = None
            if widget is self.inside:
                widget.ui_button_disarm()
                widget.invoke()

    def handle_space(self, event):
        widget = event.widget
        if widget.cget("state") != "disabled":
            # simulate a mouse click
            widget.ui_button_arm()
            widget.update_idletasks()
            widget.after(100)
            widget.ui_button_disarm()
            widget.invoke()

class ButtonMixin:
    # this mixin defines a couple of functions required
    # by the generic button binding

    ui_option_state = "normal"

    ui_option_command = None

    ui_controller = ButtonController

    def invoke(self):
        "Invoke button command"
        command = self.ui_option_command
        if callable(command):
            command()

    def ui_button_arm(self):
        # internal: mouse button pressed in button
        pass

    def ui_button_disarm(self):
        # internal: mouse button released outside button
        pass

    def ui_button_enter(self):
        # internal: mouse pointer moved into button
        pass

    def ui_button_leave(self):
        # internal: mouse pointer moved out from button
        pass

# --------------------------------------------------------------------
# scroll mixin.  implements vertical and horizontal scrolling


class ScrollMixin:

    # mixin user must override these to enable scrolling
    ui_option_xscrollcommand = tk3.NOT_INHERITED
    ui_option_yscrollcommand = tk3.NOT_INHERITED

    def ui_scroll_xinfo(self):
        # get total/left/right indexes
        raise NotImplementedError, "must override ui_scroll_xinfo"

    def ui_scroll_yinfo(self):
        # get total/top/bottom indexes
        raise NotImplementedError, "must override ui_scroll_yinfo"

    def ui_scroll_xset(self, left, units, pages):
        # set left to left + units + pages
        raise NotImplementedError, "must override ui_scroll_xset"

    def ui_scroll_yset(self, top, units, pages):
        # set top to top + units + pages
        raise NotImplementedError, "must override ui_scroll_yset"

    def ui_scroll_update(self):
        # update scrollbar
        xscrollcommand = self.ui_option_xscrollcommand
        if callable(xscrollcommand):
            total, left, right = self.ui_scroll_xinfo()
            if total > 0:
                ratio1 = float(left) / total
                ratio2 = float(right) / total
                xscrollcommand(max(0.0, ratio1), min(ratio2, 1.0))
            else:
                xscrollcommand(0.0, 1.0)
        yscrollcommand = self.ui_option_yscrollcommand
        if callable(yscrollcommand):
            total, top, bottom = self.ui_scroll_yinfo()
            if total > 0:
                ratio1 = float(top) / total
                ratio2 = float(bottom) / total
                yscrollcommand(max(0.0, ratio1), min(ratio2, 1.0))
            else:
                yscrollcommand(0.0, 1.0)

    def xview(self, command, value, unit=None):
        "Change horizontal view"
        total, left, right = self.ui_scroll_xinfo()
        if command == "moveto":
            self.ui_scroll_xset(int(total * float(value) + 0.5), 0, 0)
        elif command == "scroll":
            if unit == "units":
                self.ui_scroll_xset(left, int(value), 0)
            elif unit == "pages":
                self.ui_scroll_xset(left, 0, int(value))
        else:
            raise ValueError, "command should be 'moveto' or 'scroll'"

    def yview(self, command, value, unit=None):
        "Change vertical view"
        total, top, bottom = self.ui_scroll_yinfo()
        if command == "moveto":
            self.ui_scroll_yset(int(total * float(value) + 0.5), 0, 0)
        elif command == "scroll":
            if unit == "units":
                self.ui_scroll_yset(top, int(value), 0)
            elif unit == "pages":
                self.ui_scroll_yset(top, 0, int(value))
        else:
            raise ValueError, "command should be 'moveto' or 'scroll'"

# --------------------------------------------------------------------
# convenience widgets

class SimpleWidget(ScrollMixin, EventMixin, tk3.Widget):
    # simple scrolled event-handling widget
    pass

class ButtonWidget(ButtonMixin, tk3.Widget):
    # button widget

    ui_option_relief = "raised"
    ui_option_borderwidth = 2

    def ui_button_arm(self):
        self.__relief = self.cget("relief")
        self.config(relief="sunken")

    def ui_button_disarm(self):
        self.config(relief=self.__relief)
